pub mod gecos_fields;
pub mod passwd_fields;
pub mod shadow_fields;

use std::convert::TryFrom;
use std::fmt::{self, Display};

/// A record(line) in the user database `/etc/passwd` found in most linux systems.
#[derive(Debug, PartialEq, Eq)]
pub struct User<'a> {
    source: &'a str,
    username: crate::Username<'a>,    /* Username.  */
    password: crate::Password<'a>, /* Hashed passphrase, if shadow database not in use (see shadow.h).  */
    uid: crate::Uid,               /* User ID.  */
    gid: crate::Gid,               /* Group ID.  */
    gecos: crate::Gecos<'a>,       /* Real name.  */
    home_dir: crate::HomeDir<'a>,  /* Home directory.  */
    shell_path: crate::ShellPath<'a>, /* Shell program.  */
}

impl<'a> User<'a> {
    /// Parse a line formatted like one in `/etc/passwd` and construct a matching [`adduser::User`] instance
    ///
    /// # Example
    /// ```
    /// let pwd = adduser::User::new_from_string(
    ///     "testuser:testpassword:1001:1001:full Name,,,,:/home/test:/bin/test"
    /// ).unwrap();
    /// assert_eq!(pwd.get_username(), "testuser");
    /// ```
    ///
    /// # Errors
    /// When parsing fails this function returns a `UserLibError::Message` containing some information as to why the function failed.
    pub fn new_from_string(line: &'a str) -> Result<Self, crate::UserLibError> {
        let elements: Vec<&str> = line.split(':').collect();
        if elements.len() == 7 {
            Ok(Self {
                source: line,
                username: crate::Username::try_from(*elements.get(0).unwrap())?,
                password: crate::Password::Encrypted(crate::EncryptedPassword::try_from(
                    *elements.get(1).unwrap(),
                )?),
                uid: crate::Uid::try_from(*elements.get(2).unwrap())?,
                gid: crate::Gid::try_from(*elements.get(3).unwrap())?,
                gecos: crate::Gecos::try_from(*elements.get(4).unwrap())?,
                home_dir: crate::HomeDir::try_from(*elements.get(5).unwrap())?,
                shell_path: crate::ShellPath::try_from(*elements.get(6).unwrap())?,
            })
        } else {
            Err("Failed to parse: not enough elements".into())
        }
    }
    #[must_use]
    pub const fn get_username(&self) -> &'a str {
        self.username.username
    }
    #[must_use]
    pub const fn get_password(&self) -> &'a str {
        match self.password {
            crate::Password::Encrypted(crate::EncryptedPassword { password }) => password,
            crate::Password::Shadow(crate::Shadow { ref password, .. }) => password.password,
            crate::Password::Disabled => "x",
        }
    }
    #[must_use]
    pub const fn get_uid(&self) -> u32 {
        self.uid.uid
    }
    #[must_use]
    pub const fn get_gid(&self) -> u32 {
        self.gid.gid
    }
    #[must_use]
    pub const fn get_comment(&self) -> &crate::Gecos {
        &self.gecos
    }
    #[must_use]
    pub const fn get_home_dir(&self) -> &'a str {
        self.home_dir.dir
    }
    #[must_use]
    pub const fn get_shell_path(&self) -> &'a str {
        self.shell_path.shell
    }
}

impl Default for User<'_> {
    fn default() -> Self {
        Self {
            source: "",
            username: crate::Username {
                username: "defaultuser",
            },
            password: crate::Password::Encrypted(crate::EncryptedPassword {
                password: "notencrypted",
            }),
            uid: crate::Uid { uid: 1001 },
            gid: crate::Gid { gid: 1001 },
            gecos: crate::Gecos::Simple {
                comment: "gecos default comment",
            },
            home_dir: crate::HomeDir {
                dir: "/home/default",
            },
            shell_path: crate::ShellPath { shell: "/bin/bash" },
        }
    }
}

impl Display for User<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(
            f,
            "{}:{}:{}:{}:{}:{}:{}",
            self.username,
            self.password,
            self.uid,
            self.gid,
            self.gecos,
            self.home_dir,
            self.shell_path
        )
    }
}

#[test]
fn test_default_user() {
    // Check if a user can be created.
    let pwd = User::default();
    assert_eq!(pwd.username.username, "defaultuser");
    assert_eq!(pwd.home_dir.dir, "/home/default");
    assert_eq!(pwd.uid.uid, 1001);
}

#[test]
fn test_new_from_string() {
    // Test if a single line can be parsed and if the resulting struct is populated correctly.
    let fail = User::new_from_string("").err().unwrap();
    assert_eq!(
        fail,
        crate::UserLibError::Message("Failed to parse: not enough elements".into())
    );
    let pwd =
        User::new_from_string("testuser:testpassword:1001:1001:testcomment:/home/test:/bin/test")
            .unwrap();
    let pwd2 =
        User::new_from_string("testuser:testpassword:1001:1001:full Name,004,000342,001-2312,myemail@test.com:/home/test:/bin/test")
            .unwrap();
    assert_eq!(pwd.username.username, "testuser");
    assert_eq!(pwd.home_dir.dir, "/home/test");
    assert_eq!(pwd.uid.uid, 1001);
    match pwd.gecos {
        crate::Gecos::Simple { comment } => assert_eq!(comment, "testcomment"),
        _ => unreachable!(),
    }
    match pwd2.gecos {
        crate::Gecos::Detail {
            full_name,
            room,
            phone_work,
            phone_home,
            other,
        } => {
            assert_eq!(full_name, "full Name");
            assert_eq!(room, "004");
            assert_eq!(phone_work, "000342");
            assert_eq!(phone_home, "001-2312");
            assert_eq!(other.unwrap()[0], "myemail@test.com");
        }
        _ => unreachable!(),
    }
}

#[test]
fn test_parse_passwd() {
    // Test wether the passwd file can be parsed and recreated without throwing an exception
    use std::fs::File;
    use std::io::{prelude::*, BufReader};
    let file = File::open("/etc/passwd").unwrap();
    let reader = BufReader::new(file);

    for line in reader.lines() {
        let lineorig: String = line.unwrap();
        let linecopy = lineorig.clone();
        let pass_struc = User::new_from_string(&linecopy).unwrap();
        assert_eq!(
            // ignoring the numbers of `,` since the implementation does not (yet) reproduce a missing comment field.
            lineorig,
            format!("{}", pass_struc)
        );
    }
}
