#![warn(
    clippy::all,
/*    clippy::restriction,*/
    clippy::pedantic,
    clippy::nursery,
    clippy::cargo
)]
#![allow(clippy::non_ascii_literal)]

use log::warn;
use regex::Regex;

use crate::userlib_error::UserLibError;
use std::cmp::Eq;
use std::convert::TryFrom;
use std::fmt::{self, Display};

/// The username of the current user
///
/// When done the validity will automatically be checked in the `trait TryFrom`.
///
/// In the future some extra fields might be added.
#[derive(Debug, PartialEq, Eq)]
pub struct Username<'a> {
    /// The username value
    pub(in crate::user) username: &'a str,
}

impl Display for Username<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.username,)
    }
}

impl<'a> TryFrom<&'a str> for Username<'a> {
    type Error = UserLibError;
    fn try_from(source: &'a str) -> std::result::Result<Self, Self::Error> {
        lazy_static! {
            static ref USERVALIDATION: Regex =
                Regex::new("^[a-z_]([a-z0-9_\\-]{0,31}|[a-z0-9_\\-]{0,30}\\$)$").unwrap();
        }
        if USERVALIDATION.is_match(source) {
            Ok(Self { username: source })
        } else if source == "Debian-exim" {
            warn!("username {} is not a valid username. This might cause problems. (It is default in Debian and Ubuntu)", source);
            Ok(Self { username: source })
        } else {
            Err(UserLibError::Message(format!(
                "Invalid username {}",
                source
            )))
        }
    }
}

#[derive(Debug, PartialEq, Eq)]
pub enum Password<'a> {
    Encrypted(crate::EncryptedPassword<'a>),
    Shadow(crate::Shadow<'a>),
    Disabled,
}

impl Display for Password<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Password::Encrypted(EncryptedPassword { password }) => write!(f, "{}", password,),
            Password::Shadow(_) => write!(f, "x"),
            Password::Disabled => write!(f, "x"),
        }
    }
}

#[derive(Debug, PartialEq, Eq)]
pub struct EncryptedPassword<'a> {
    pub(in crate::user) password: &'a str,
}

impl Display for EncryptedPassword<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.password,)
    }
}

impl<'a> TryFrom<&'a str> for EncryptedPassword<'a> {
    type Error = UserLibError;
    fn try_from(source: &'a str) -> std::result::Result<Self, Self::Error> {
        if source == "x" {
            warn!("password from shadow not loaded!")
        } else {
            warn!("Password field has an unexpected value")
        };
        Ok(Self { password: source })
    }
}

#[derive(Debug, PartialEq, Eq)]
pub struct Uid {
    pub(in crate::user) uid: u32,
}

impl Display for Uid {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.uid,)
    }
}

impl TryFrom<&str> for Uid {
    type Error = UserLibError;
    fn try_from(source: &str) -> std::result::Result<Self, Self::Error> {
        Ok(Self {
            uid: source.parse::<u32>().unwrap(),
        })
    }
}

impl Uid {
    #[must_use]
    pub const fn is_system_uid(&self) -> bool {
        // since it is a u32  it cannot be smaller than 0
        self.uid < 1000
    }
}

#[derive(Debug, PartialEq, Eq)]
pub struct Gid {
    pub(in crate::user) gid: u32,
}

impl Display for Gid {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.gid,)
    }
}

impl TryFrom<&str> for Gid {
    type Error = UserLibError;
    fn try_from(source: &str) -> std::result::Result<Self, Self::Error> {
        Ok(Self {
            gid: source.parse::<u32>().unwrap(),
        })
    }
}

impl Gid {
    #[must_use]
    pub const fn is_system_gid(&self) -> bool {
        // since it is a u32  it cannot be smaller than 0
        self.gid < 1000
    }
}

/// The home directory of a user
#[derive(Debug, PartialEq, Eq)]
pub struct HomeDir<'a> {
    pub(in crate::user) dir: &'a str,
}

impl Display for HomeDir<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.dir,)
    }
}

impl<'a> TryFrom<&'a str> for HomeDir<'a> {
    type Error = UserLibError;
    fn try_from(source: &'a str) -> std::result::Result<Self, Self::Error> {
        Ok(Self { dir: source })
    }
}

/// The path to the Shell binary
#[derive(Debug, PartialEq, Eq)]
pub struct ShellPath<'a> {
    pub(in crate::user) shell: &'a str,
}

impl Display for ShellPath<'_> {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        write!(f, "{}", self.shell,)
    }
}

impl<'a> TryFrom<&'a str> for ShellPath<'a> {
    type Error = UserLibError;
    fn try_from(source: &'a str) -> std::result::Result<Self, Self::Error> {
        Ok(ShellPath { shell: source })
    }
}

// Tests ----------------------------------------------------------------------

#[test]
fn test_username_validation() {
    // Failing tests
    let umlauts = Username::try_from("täst"); // umlauts
    assert_eq!(
        Err(UserLibError::Message("Invalid username täst".into())),
        umlauts
    );
    let number_first = Username::try_from("11elf"); // numbers first
    assert_eq!(
        Err(UserLibError::Message("Invalid username 11elf".into())),
        number_first
    );
    let slashes = Username::try_from("test/name"); // slashes in the name
    assert_eq!(
        Err(UserLibError::Message("Invalid username test/name".into())),
        slashes
    );
    let long = Username::try_from("aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"); // maximum size 32 letters
    assert_eq!(
        Err(UserLibError::Message(
            "Invalid username aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa".into()
        )),
        long
    );
    // Working tests
    let ubuntu_exception = Username::try_from("Debian-exim"); // for some reason ubuntu and debian have a capital user.
    assert_eq!(ubuntu_exception.unwrap().username, "Debian-exim");
    let single = Username::try_from("t"); // single characters are ok
    assert_eq!(single.unwrap().username, "t");
    let normal = Username::try_from("superman"); // regular username
    assert_eq!(normal.unwrap().username, "superman");
    let normal = Username::try_from("anna3pete"); // regular username containing a number
    assert_eq!(normal.unwrap().username, "anna3pete");
    let normal = Username::try_from("enya$"); // regular username ending in a $
    assert_eq!(normal.unwrap().username, "enya$");
}

#[test]
fn test_guid_system_user() {
    // Check uids of system users.
    let values = vec![("999", true), ("0", true), ("1000", false)];
    for val in values {
        assert_eq!(Uid::try_from(val.0).unwrap().is_system_uid(), val.1);
        assert_eq!(Gid::try_from(val.0).unwrap().is_system_gid(), val.1);
    }
}
